// $Id: qvis.cxx,v 1.3 1997/09/15 13:27:11 garland Exp $

#include <gfx/std.h>
#include <gfx/XF/glxf.h>

void gui_init(int *, char **);
void gui_loop();

#define GUI
#include "./qslim.cxx"

bool will_draw_mesh = false;
bool will_draw_surface = true;
bool will_draw_vquadrics = false;
bool will_draw_equadrics = false;
bool will_smooth_surface = false;
bool will_shade_by_error = false;
bool two_sided_lighting = false;

bool ntsc_output = false;

real ellipsoid_radius = 1.0;



#include <strstream.h>
#include <fstream.h>
#include <gfx/GL/utils.h>
#include <gfx/GL/Camera.h>
#include <gfx/manip/Arcball.h>
#include "VisQuadrics.h"

extern "C"
{
#include "qvis_form.h"
}

static FD_view_panel *UI;
static glCamera camera;
static Arcball manip;
static GLUquadricObj *quadric;

static
void update_status(char *msg=NULL)
{
    ostrstream out;

    if( msg )
	fl_set_input(UI->status_line, msg);
    else if( ntsc_output )
    {
	out << "Faces: " << M0.validFaceCount << " / " << initialFaceCount;
	char *tmp = out.str();
	fl_set_input(UI->status_line, tmp);
	delete tmp;
    }
    else
    {
	out << "Original: "
	    << initialVertCount << "v/"
	    << initialEdgeCount << "e/"
	    << initialFaceCount << "f"
	    << "        "
	    << "Current: "
	    << M0.validVertCount << "v/"
	    << M0.validEdgeCount << "e/"
	    << M0.validFaceCount << "f";

	char *tmp = out.str();
	fl_set_input(UI->status_line, tmp);
	delete tmp;
    }
}

//
// Set to true when the model has changed.  This will force update of
// any cached rendering state (eg. ellipsoid dlists)
//
static bool model_dirty_bit = true;

static
void draw_vertex_ellipsoids(Model& m, real rad=1.0)
{
    glPushAttrib(GL_LIGHTING_BIT);
    glDisable(GL_COLOR_MATERIAL);
    lightEllipsoids(ELLIPSOID_GREEN);

    for(int i=0; i<m.vertCount(); i++)
	if( m.vertex(i)->isValid() )
	{
	    Vertex *v = m.vertex(i);
	    Mat4 Q;

	    if( decimate_quadric(v, Q) )
		drawEllipsoid(quadric, Q, rad, v);
	}
    glPopAttrib();
}

static
void draw_edge_ellipsoids(Model& m, real rad=0.5)
{
    glPushAttrib(GL_LIGHTING_BIT);
    glDisable(GL_COLOR_MATERIAL);
    lightEllipsoids(ELLIPSOID_RED);

    for(int i=0; i<m.edgeCount(); i++)
	if( m.edge(i)->isValid() )
	{
	    Vertex *v1 = m.edge(i)->org();
	    Vertex *v2 = m.edge(i)->dest();
	    Mat4 Q1, Q2;

	    if( decimate_quadric(v1, Q1) && decimate_quadric(v2, Q2) )
		drawEllipsoid(quadric, Q1+Q2, rad, v1);
	}
    glPopAttrib();
}

static real current_max_error = 0.0;

static
void error_color(Vertex& v)
{
    real err = decimate_error(&v);
    real val = (err)/(current_max_error);

    glC(val, val, val);
}

static
void draw_surface(Model& m)
{
    if( will_shade_by_error )
    {
	current_max_error = decimate_max_error(m);
    }
    glBegin(GL_TRIANGLES);
    for(int i=0; i<m.faceCount(); i++)
	if( m.face(i)->isValid() )
	{
	    Face& f = *m.face(i);
	    Vertex& v1 = *f.vertex(0);
	    Vertex& v2 = *f.vertex(1);
	    Vertex& v3 = *f.vertex(2);
	    
	    glN(f.plane().normal());
#ifdef SUPPORT_FCOLOR
	    if( f.props )
		glC(f.props->color);
	    else
		glC(0.5, 0.5, 0.5);
#endif
#ifdef SUPPORT_VCOLOR
	    if( v1.props ) glC(v1.props->color);
#endif
	    if( will_smooth_surface ) glN(m.synthesizeNormal(&v1));
	    if( will_shade_by_error ) error_color(v1);
	    glV(v1);
#ifdef SUPPORT_VCOLOR
	    if( v2.props ) glC(v2.props->color);
#endif
	    if( will_smooth_surface ) glN(m.synthesizeNormal(&v2));
	    if( will_shade_by_error ) error_color(v2);
	    glV(v2);
#ifdef SUPPORT_VCOLOR
	    if( v3.props ) glC(v3.props->color);
#endif
	    if( will_smooth_surface ) glN(m.synthesizeNormal(&v3));
	    if( will_shade_by_error ) error_color(v3);
	    glV(v3);
	}
    glEnd();
}

static
void draw_mesh(Model& m)
{
    glPushAttrib(GL_LIGHTING_BIT);
    glDisable(GL_LIGHTING);
    glC(0.8, 0.1, 0.1);

    glBegin(GL_LINES);
    for(int i=0; i<m.edgeCount(); i++)
	if( m.edge(i)->isValid() )
	{
	    Edge& e = *m.edge(i);
	    Vertex& v1 = *e.org();
	    Vertex& v2 = *e.dest();
		
	    glV(v1);
	    glV(v2);
	}
    glEnd();

    glPopAttrib();
}

static
void redraw()
{
    Model& m = M0;

    glxf_make_current(UI->view);
    glClearColor(0.3, 0.3, 1.0, 0);
    //glClearColor(1.0, 1.0, 1.0, 0);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    manip.apply();
 
    glC(0.5, 0.5, 0.5); // Enforce the default color

    if( will_draw_surface )
	draw_surface(m);

    if( will_draw_mesh )
	draw_mesh(m);

    static GLuint vq_dlist = 3, eq_dlist = 4;
    static bool vq_invalid = true, eq_invalid = true;

    if( will_draw_vquadrics )
	if( model_dirty_bit || vq_invalid )
	{
	    glNewList(vq_dlist, GL_COMPILE_AND_EXECUTE);
	    draw_vertex_ellipsoids(m, ellipsoid_radius);
	    glEndList();
	    vq_invalid = false;
	}
	else
	    glCallList(vq_dlist);

    if( will_draw_equadrics )
	if( model_dirty_bit || eq_invalid )
	{
	    glNewList(eq_dlist, GL_COMPILE_AND_EXECUTE);
	    draw_edge_ellipsoids(m, ellipsoid_radius);
	    glEndList();
	    eq_invalid = false;
	}
	else
	    glCallList(eq_dlist);

    manip.unapply();
    glxf_finish(UI->view);
    update_status();
    model_dirty_bit = false;
}

void cb_decimate(FL_OBJECT *, long options)
{
    static bool initialized = false;
    bool will_redisplay = true;

    if( !initialized )
    {
	decimate_init(M0, pair_selection_tolerance);
	initialized = true;
    }

    if( !options )
	will_redisplay = false;

    face_target = atoi(fl_get_input(UI->face_target));

    if( options < 0 )
    {
	decimate_contract(M0);
	model_dirty_bit = true;
	redraw();
	return;
    }

    while( M0.validFaceCount > face_target &&
	   decimate_min_error() < error_tolerance )
    {
	decimate_contract(M0);

	if( will_redisplay )
	{
	    model_dirty_bit = true;
	    redraw();
	    fl_check_only_forms();
	}
    }

    model_dirty_bit = true;
    redraw();
}

void cb_save_model(FL_OBJECT *, long)
{
    if( !outfile )
    {
	const char *filename = fl_show_fselector("Save model to file:",
						 ".",
						 "*.smf",
						 "out.smf");
	if( filename )
	    outfile = new ofstream(filename);
    }

    if( outfile )
	*outfile << M0;
}

void cb_snapshot(FL_OBJECT *, long)
{
    const char *filename = fl_show_fselector("Snapshot to file:", ".", "*.ppm",
                                             "out.ppm");

    if( filename )
    {
        redraw();
        fl_check_only_forms();
        update_status("Saving displayed image ...");
        fl_check_only_forms();

        ofstream ppm(filename);
        glToPPM(ppm);
        ppm.close();

        update_status();
    }
}

static
int cb_key(FL_OBJECT *, Window, int, int, XEvent *event, void *)
{
    char c;
    int i;

    XLookupString(&event->xkey, &c, 1, NULL, NULL);

    switch( c )
    {
    case 'q':
	exit(0);
	break;

    case 'm':
	will_draw_mesh = !will_draw_mesh;

	if( will_draw_mesh )
	    glOffsetForMesh();
	else
	    glNoOffset();

	break;

    case 'n':
	ntsc_output = !ntsc_output;
	if( ntsc_output )
	{
	    glEnable(GL_LINE_SMOOTH);
	    glLineWidth(1.1);
	}
	else
	{
	    glDisable(GL_LINE_SMOOTH);
	    glLineWidth(1.0);
	}
	break;

    case 'p':
	will_shade_by_error = !will_shade_by_error;
	if( will_shade_by_error )
	    glDisable(GL_LIGHTING);
	else
	    glEnable(GL_LIGHTING);
	break;

    case 's':
	will_draw_surface = !will_draw_surface;
	break;

    case 'e':
	will_draw_vquadrics = !will_draw_vquadrics;
	break;

    case 'E':
	will_draw_equadrics = !will_draw_equadrics;
	break;

    case 'g':
	will_smooth_surface = !will_smooth_surface;
	break;

    case 'd':
	two_sided_lighting = !two_sided_lighting;
	glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, two_sided_lighting);
	break;

    case 'v':
	cerr << "Updating valid info ..." << endl;
	M0.validVertCount = 0;
	M0.validEdgeCount = 0;
	M0.validFaceCount = 0;

	for(i=0; i<M0.vertCount(); i++)
	    if( M0.vertex(i)->isValid() )
		M0.validVertCount++;

	for(i=0; i<M0.edgeCount(); i++)
	    if( M0.edge(i)->isValid() )
		M0.validEdgeCount++;

	for(i=0; i<M0.faceCount(); i++)
	    if( M0.face(i)->isValid() )
		M0.validFaceCount++;

	update_status();
	return 0;
	// break;

    case ',':
	ellipsoid_radius *= 0.5;
	model_dirty_bit = true;
	break;
    case '.':
	ellipsoid_radius *= 2;
	model_dirty_bit = true;
	break;

    default:
	return 0;
    }

    redraw();
    return 0;
}

static
int cb_button(FL_OBJECT *, Window, int, int, XEvent *event, void *)
{
    if( manip.process_event(*event) )
        redraw();

    return 0;
}

static
int cb_expose(FL_OBJECT *canvas, Window,
              int width, int height, XEvent *, void *)
{
    if( !canvas->visible ) return 0;

    glViewport(0, 0, width, height);
    redraw();
    return 0;
}

static
int cb_mapnotify(FL_OBJECT *canvas, Window, int, int, XEvent *, void *)
{
    glxf_make_current(canvas);

    glEnable(GL_DEPTH_TEST);
    glEnable(GL_LIGHTING);
    glEnable(GL_NORMALIZE);
    if( ntsc_output )
    {
	glEnable(GL_LINE_SMOOTH);
	glLineWidth(1.1);
    }
    glShadeModel(GL_SMOOTH);

    glMatrixMode(GL_PROJECTION);
    camera.glApply();

    glMatrixMode(GL_MODELVIEW);

    GLfloat mat_ambient[] = { 0.1, 0.1, 0.1, 1.0 };
    GLfloat mat_diffuse[] = { 0.5, 0.5, 0.5, 1.0 };
    glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, mat_ambient);
    glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, mat_diffuse);

    GLfloat light_pos[] = {camera.from[X], camera.from[Y], camera.from[Z]};
    glLightfv(GL_LIGHT0, GL_POSITION, light_pos);
    glEnable(GL_LIGHT0);

    glColorMaterial(GL_FRONT_AND_BACK, GL_DIFFUSE);
    glEnable(GL_COLOR_MATERIAL);
    glC(0.5, 0.5, 0.5);

    quadric = newEllipsoidContext();

    return 0;
}

void gui_init(int *argc, char **argv)
{
    fl_initialize(argc, argv, "QSlim/Vis", 0, 0);
    glxf_set_defaults(GLXF_CFG_DOUBLEBUFFER);
}

void gui_create(Model& m)
{
    UI = create_form_view_panel();

    //
    // Setup UI defaults
    //
    fl_deactivate_object(UI->status_line);
    ostrstream out;
    out << face_target;
    fl_set_input(UI->face_target, out.str());
    delete out.str();

    fl_add_canvas_handler(UI->view, Expose, cb_expose, NULL);
    fl_add_canvas_handler(UI->view, MapNotify, cb_mapnotify, NULL);
    fl_add_canvas_handler(UI->view, ButtonPress, cb_button, NULL);
    fl_add_canvas_handler(UI->view, ButtonRelease, cb_button, NULL);
    fl_add_canvas_handler(UI->view, MotionNotify, cb_button, NULL);
    fl_add_canvas_handler(UI->view, KeyPress, cb_key, NULL);

    //
    // Display form and finish initialization
    //
    fl_set_form_position(UI->view_panel, 0, 0);
    fl_show_form(UI->view_panel, FL_PLACE_CENTER, FL_FULLBORDER, "QSlim/Vis");

    camera.lookAt(m.bounds, 640.0f/480.0f);
    manip.init(m.bounds, 640.0f/480.0f);
}

void gui_loop()
{
    gui_create(M0);

    while( true )
	fl_do_forms();
}
